\ FIND files over all drives and all users
\ Source text for DX-Forth
\ ZCPR3 compliant
\ Pavel Zampach, 2020

true \ for standard CP/M or changed RunCPM, '?' in FCB allowed
\ false \ for original RunCPM

decimal
16  value #USR      \ no of users
16  value #DRV      \ no of drives
$80 constant DEFDMA \ default dma
$5C constant DEFFCB \ default fcb
variable ORIDRV     \ original drive
variable ORIUSR     \ original user

\ addr of NDR buffer
: NDR-ADDR ( -- addr )
  zenv dup if 21 + @ then
;

\ ndr item addr -> dir name (addr-c u)
: ADR>NAME ( addr -- addr-c u )
  2+ 8 -trailing
;

\ addr of next NDR item
: NDR-NEXT ( addr1 -- addr2 )
  18 +
;

\ search for disk (A = 0) and user, returns dir name (addr-c u)
\ if nothing was found or NDR is not implemented, returns 0
: NDR-NAME ( disk user -- addr-c u | false)
  >< or 1+       \ high byte = user, low byte = disk
  ndr-addr       \ du addr-ndr -- ;
  dup if
    begin
      dup c@     \ end?
    while
      2dup @ = if
        nip adr>name
      end
      ndr-next
    repeat
  then
  2drop 0        \ not found
;

\ print dir name (if exists) in brackets
: .NDR-NAME ( disk user -- )
  ndr-name ?dup if
    [char] / emit
    type
\    [char] ) emit
  then
;

\ print stack w/o trailing space
: .USER
  s>d <# #s #> type
;

\ print complete filename
\ addr = dir entry, disk = 0 for A...
: .FILENAME ( addr disk user -- )
  over [char] A + emit    \ disk char
  dup .user               \ user
  .ndr-name               \ dir name
  [char] : emit           \ colon
  13 out @ - spaces       \ align by spaces
  dup 1+ 8 -trailing type \ filename
  [char] . emit           \ dot
  9 + 3 type              \ filetype
  cr
;

: SETDRV ( drive -- )
  14 bdos drop
;

: GETDRV ( -- drive )
  0 25 bdos
;

\ loop for searching all drives
[IF] \ standard CP/M

: DRV-LOOP
  cr
  defdma 26 bdos drop  \ set DMA
  #drv 0 do
    i setdrv           \ set drive
    [char] ? deffcb c! \ search over all users
    deffcb 17 bdos     \ first search
    begin
      dup $FF <>       \ file exists?
    while
      32 * defdma + i over c@ .filename
      0 18 bdos        \ next file
    repeat   drop
  loop
;

[ELSE] \ original RunCPM

: DRV-LOOP
  cr
  defdma 26 bdos drop  \ set DMA
  #drv 0 do            \ drive loop
    i setdrv           \ set drive
    #usr 0 do          \ user loop
      i setusr         \ set user
      deffcb 17 bdos   \ first search
      begin
        dup $FF <>     \ file exists?
      while
        32 * defdma + j i .filename
        0 18 bdos      \ next file
      repeat   drop
    loop
  loop
;

[THEN]

: MAIN
  defdma c@ 0= abort" Usage: FIND afn"
  zenv ?dup if
    44 + dup c@ to #drv
    1+ c@ 1+ to #usr
  then
  getdrv oridrv ! \ save original drive
  getusr oriusr ! \ save original user
  drv-loop
  oriusr @ setusr \ restore original user
  oridrv @ setdrv \ restore original drive
;
